/*
 * characters in village
 *
 * Copyright(c) 2008 olyutorskii
 * $Id: Avatar.java 487 2009-04-25 12:28:20Z olyutorskii $
 */

package jp.sourceforge.jindolf;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import jp.sourceforge.jindolf.core.PreDefAvatar;

/**
 * Avatar またの名をキャラクター
 */
public class Avatar implements Comparable<Avatar>{

    private static final List<PreDefAvatar> preDefAvatarList =
            PreDefAvatar.getPreDefAvatarList();
    private static final SortedSet<Avatar> predefinedAvatarSet;
    private static final List<Avatar> predefinedAvatarList;
    private static final Map<String, Avatar> avatarMap;

    private static final Pattern avatarRegex;

    public static final Avatar AVATAR_GERD;

    static{
        predefinedAvatarSet = loadPredefSet();

        predefinedAvatarList =
                new ArrayList<Avatar>(predefinedAvatarSet.size());
        avatarMap = new TreeMap<String, Avatar>();
        String avatarGroup = "";
        for(Avatar avatar : predefinedAvatarSet){
            predefinedAvatarList.add(avatar);

            String fullName = avatar.getFullName();

            avatarMap.put(fullName, avatar);

            if(avatarGroup.length() > 0){
                avatarGroup += "|";
            }
            avatarGroup += "(" + Pattern.quote(fullName) + ")";
        }

        avatarRegex = Pattern.compile(avatarGroup);

        AVATAR_GERD = getPredefinedAvatar("楽天家 ゲルト");
    }

    /**
     * 定義済みAvatarのロード。
     */
    private static SortedSet<Avatar> loadPredefSet(){
        SortedSet<Avatar> result = new TreeSet<Avatar>();

        for(PreDefAvatar preDefAvatar : preDefAvatarList){
            String shortName = preDefAvatar.getShortName();
            String jobTitle = preDefAvatar.getJobTitle();
            int serialNo = preDefAvatar.getSerialNo();
            String avatarId = preDefAvatar.getAvatarId();
            Avatar avatar = new Avatar(shortName, jobTitle, serialNo, avatarId);
            result.add(avatar);
        }

        result = Collections.unmodifiableSortedSet(result);

        return result;
    }

    /**
     * 定義済みAvatar群のリストを返す。
     * @return Avatarのリスト
     */
    public static SortedSet<Avatar> getPredefinedAvatarSet(){
        return predefinedAvatarSet;
    }

    /**
     * 定義済みAvatarを返す。
     * @param fullName Avatarのフルネーム
     * @return Avatar
     */
    public static Avatar getPredefinedAvatar(String fullName){
        return avatarMap.get(fullName);
    }

    /**
     * 定義済みAvatar名に一致しないか調べる。
     * @param matcher マッチャ
     * @return 一致したAvatar。一致しなければnull。
     */
    public static Avatar lookingAtAvatar(Matcher matcher){
        matcher.usePattern(avatarRegex);

        if( ! matcher.lookingAt() ) return null;
        int groupCt = matcher.groupCount();
        for(int group = 1; group <= groupCt; group++){
            if(matcher.start(group) >= 0){
                Avatar avatar = predefinedAvatarList.get(group-1);
                return avatar;
            }
        }

        return null;
    }

    private final String name;
    private final String jobTitle;
    private final String fullName;
    private final int idNum;
    private final String identifier;

    /**
     * Avatarを生成する。
     * @param name 名前
     * @param jobTitle 職業名
     * @param idNum 通し番号
     * @param identifier 識別文字列
     */
    private Avatar(String name,
                    String jobTitle,
                    int idNum,
                    String identifier ){
        this.name = name.intern();
        this.jobTitle = jobTitle.intern();
        this.idNum = idNum;
        this.identifier = identifier.intern();

        this.fullName = (this.jobTitle + " " + this.name).intern();

        return;
    }

    /**
     * Avatarを生成する
     * @param fullName フルネーム
     */
    // TODO 当面は呼ばれないはず。Z国とか向け。
    public Avatar(String fullName){
        this.fullName = fullName.intern();
        this.idNum = -1;

        String[] tokens = this.fullName.split("\\p{Blank}+", 2);
        if(tokens.length == 1){
            this.jobTitle = null;
            this.name = this.fullName;
        }else if(tokens.length == 2){
            this.jobTitle = tokens[0].intern();
            this.name = tokens[1].intern();
        }else{
            this.jobTitle = null;
            this.name = null;
            assert false;
        }

        this.identifier = "???".intern();

        return;
    }

    /**
     * フルネームを取得する。
     * @return フルネーム
     */
    public String getFullName(){
        return this.fullName;
    }

    /**
     * 職業名を取得する。
     * @return 職業名
     */
    public String getJobTitle(){
        return this.jobTitle;
    }

    /**
     * 通常名を取得する。
     * @return 通常名
     */
    public String getName(){
        return this.name;
    }

    /**
     * 通し番号を返す。
     * @return 通し番号
     */
    public int getIdNum(){
        return this.idNum;
    }

    /**
     * 識別文字列を返す。
     * @return 識別文字列
     */
    public String getIdentifier(){
        return this.identifier;
    }

    /**
     * {@inheritDoc}
     * @param obj {@inheritDoc}
     * @return {@inheritDoc}
     */
    @Override
    public boolean equals(Object obj){
        if(this == obj){
            return true;
        }
        if(obj == null){
            return false;
        }
        if( ! (obj instanceof Avatar) ){
            return false;
        }

        Avatar other = (Avatar) obj;

        boolean nameMatch = this.fullName.equals(other.fullName);
        boolean idMatch = this.idNum == other.idNum;

        if(nameMatch && idMatch) return true;

        return false;
    }

    /**
     * {@inheritDoc}
     * @return {@inheritDoc}
     */
    @Override
    public int hashCode(){
        return this.fullName.hashCode() ^ this.idNum;
    }

    @Override
    public String toString(){
        return getFullName();
    }

    /**
     * {@inheritDoc}
     * @param avatar {@inheritDoc}
     * @return {@inheritDoc}
     */
    public int compareTo(Avatar avatar){
        if(avatar == null) return +1;
        return this.idNum - avatar.idNum;
    }

}
